#include <iostream>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <list>
#include <math.h>
#include <fstream>
#include <iomanip>

#include "bitarray.h"
#include "sdes.h"

using namespace std;

class gArgs{
public:
    bool o_enc, o_dec, o_bru;
    std::string o_key,o_msg, o_cip, o_file;

    gArgs(){
        o_enc = false;
        o_dec = false;
        o_bru = false;
        o_key = "";
        o_msg = "";
        o_cip = "";
        o_file = "";
    }

};

std::string b_add(std::string s1){
    int i;
    int count = 0;
    std::string s[10], ret;
    std::string s2 = "0000000001";

    for(i = 9; i>=0; i--){
        if(s1[i] == '0' && s2[i] == '0'){
            if(count){
                s[i] = '1';
                count = 0;
            }
            else{
                s[i] = '0';
            }
        }

        if((s1[i] == '0' && s2[i] == '1') ||
           (s1[i] == '1' && s2[i] == '0')){
            if(count){
                s[i] = '0';
                count = 1;
            }
            else{
                s[i] = '1';
            }
        }

        if(s1[i] == '1' && s2[i] == '1'){
            if(count){
                s[i] = '1';
                count = 0;
            }
            else{
                s[i] = '0';
                count = 1;
            }
        }
    }
    for(i = 0; i<10; i++){
        ret.append(s[i]);
    }
    return ret;
}

std::string int2string(int num, int length)
{
    char buffer[32];
    itoa(num,buffer,2);

    std::string s = std::string(buffer);
    while(s.length()<(unsigned) length){
        s = "0" + s;
    }
    return s;
}

int string2int(std::string s){
     int i = s.length(),
         sum = 0;

     while(i>0){
         if(s[s.length()-i] == '1')
             sum += pow(2,i-1);
         i--;
     }
     return sum;
}



void print_usage(void){
    std::cout << "\nDescription:\n";
    std::cout << "\tsdes is a small program that encrypts/decrypts\n";
    std::cout << "\ta given input using the simple DES algorithm.\n\n";
    std::cout << "\tApostolos Kechagias ICTE-UoWM Copyleft (c) 2012\n\n";

    std::cout << "options:\n";
    std::cout << "\t -e:\t encrypt flag\n";
    std::cout << "\t -d:\t decrypt flag\n";
    std::cout << "\t -c:\t define cipher\n";
    std::cout << "\t -m:\t define message\n";
    std::cout << "\t -k:\t encrypt/decrypt key\n";
    std::cout << "\t -f:\t encrypt/decrypt specified file EXPERIMENTAL\n";
    std::cout << "\t  d:\t to use decimal numbers\n";

    std::cout << "examples:\n";
    std::cout << "\t sdes -e -m 11110011 -k 1010000010\n";
    std::cout << "\t sdes -d -c 01000001 -k 1010000010\n";
    std::cout << "\t sdes -e -f test.txt -k d100\n";
    std::cout << "\t sdes -b -m d150 -c d44\n\n";
    exit(EXIT_FAILURE);
}

int main(int argc, char** argv)
{

    static const char *optString = "edbk:c:m:f:h?";
    gArgs args;
    std::string tmp;
    if(argc <= 1) print_usage();
    int opt = 0;
    opt = getopt(argc, argv, optString);
    while(opt != -1){
        switch(opt){
        case 'e':
            args.o_enc = true;
            break;
        case 'd':
            args.o_dec = true;
            break;
        case 'b':
            args.o_bru = true;
            break;
        case 'k':
            if(optarg[0] == 'd'){
                tmp = optarg;
                int val = atoi(tmp.substr(1,tmp.length()).c_str());
                if(val>=0 && val<1024){
                    args.o_key = int2string(val,10);
                }
                else{
                    std::cout << "error: decimal key range [0-1023]!" << std::endl;
                    print_usage();
                }
            }
            else{
                args.o_key = optarg;
                if(args.o_key.length()<10){
                    std::cout << "error: 10 bits needed for key!" << std::endl;
                    print_usage();
                }
            }
            break;
        case 'c':
            if(optarg[0] == 'd'){
                tmp = optarg;
                int val = atoi(tmp.substr(1,tmp.length()).c_str());
                if(val>=0 && val<256){
                    args.o_cip = int2string(val,8);
                }
                else{
                    std::cout << "error: decimal cipher range [0-255]!" << std::endl;
                    print_usage();
                }
            }
            else{
                args.o_cip = optarg;
                if(args.o_cip.length()<8){
                    std::cout << "error: 8 bits needed for cipher!" << std::endl;
                    print_usage();
                }
            }
            break;
        case 'm':
            if(optarg[0] == 'd'){
                tmp = optarg;
                int val = atoi(tmp.substr(1,tmp.length()).c_str());
                if(val>=0 && val<256){
                    args.o_msg = int2string(val,8);
                }
                else{
                    std::cout << "error: decimal message range [0-255]!" << std::endl;
                    print_usage();
                }
            }
            else{
                args.o_msg = optarg;
                if(args.o_msg.length()<8){
                    std::cout << "error: 8 bits needed for message!" << std::endl;
                    print_usage();
                }
            }
            break;
        case 'f':
            args.o_file = optarg;
            break;
        case 'h': case '?':
            print_usage();
            break;
        default:
            print_usage();
            break;

        }
        opt = getopt(argc, argv, optString);
    }
    if((args.o_dec && args.o_enc) ||
       (args.o_dec && args.o_bru) ||
       (args.o_dec && args.o_enc) ){
        std::cout << "error: only one of -e, -d  and -b is allowed!\n\n";
        print_usage();
    }

    if((args.o_enc || args.o_dec) && args.o_file == ""){
        if(((args.o_enc && (args.o_key == "" || args.o_msg == ""))) ||
                (args.o_dec &&  (args.o_cip == "" || args.o_key == ""))){
            std::cout << "error: message/cipher and key must be specified!\n\n";
            print_usage();
        }

        BitArray text(8), key_ed(10);

        text.fromString((args.o_dec?args.o_cip:args.o_msg));
        key_ed.fromString(args.o_key);

        if(args.o_enc){
            text = _crypt(M_ENCRYPT,text,key_ed);
            std::cout << "\nMessage:\t" << args.o_msg << ":" << string2int(args.o_msg) << std::endl;
            std::cout << "Key:\t\t" << args.o_key << ":" << string2int(args.o_key) << std::endl;
            std::cout << "Cipher:\t\t" << text.toString() << ":" << string2int(text.toString()) << std::endl;
        }
        else{
            text = _crypt(M_DECRYPT,text,key_ed);
            std::cout << "\nCipher:\t\t" << args.o_cip << ":" << string2int(args.o_cip)<< std::endl;
            std::cout << "Key:\t\t" << args.o_key << ":" << string2int(args.o_key) << std::endl;
            std::cout << "Message:\t" << text.toString() << ":" << string2int(text.toString()) << std::endl;
        }
    }
    else if((args.o_enc || args.o_dec) && args.o_file != ""){
        BitArray tmp_b(8), key_edf(10);

        if(args.o_key == ""){
            std::cout << "error: a key must be specified!\n\n";
            print_usage();
        }

        key_edf.fromString(args.o_key);

        std::fstream f;
        std::fstream::pos_type pos_b, start, end;
        f.open(args.o_file.c_str(),std::ios::binary | std::ios::in | std::ios::out);

        int f_size, f_counter = 0, f_l;
        if(f.is_open()){
            start = f.tellg();
            f.seekg(0,std::ios_base::end);
            end = f.tellg();
            f.seekg(0,std::ios_base::beg);
            f_size = static_cast<int>(end-start);
            while(!f.eof()){
                pos_b = f.tellp();
                tmp_b.fromString(int2string(f.get(),8));

                if(args.o_enc){
                    tmp_b = _crypt(M_ENCRYPT,tmp_b,key_edf);
                }
                else{
                    tmp_b = _crypt(M_DECRYPT,tmp_b,key_edf);
                }
                f.seekp(pos_b);
                f.put((char)string2int(tmp_b.toString()));
                f.flush();

                f_counter++;
                float loops = (float)f_counter/f_size;
                std::cout << "\r[" ;
                for(f_l = 0; f_l<40; f_l++){
                    if(f_l<40*loops) std::cout << "#";
                    else std::cout << "-";
                }
                std::cout << "] " << setw(3) << (int)(100*loops) << "%";
            }
        }
        f.close();
    }
    else if(args.o_bru){
        if(args.o_cip == "" || args.o_msg == ""){
            std::cout << "error: message and cipher must be specified!\n\n";
            print_usage();
        }

        std::list<std::string> key_list;
        std::list<std::string>::iterator it;

        timeval t_start, t_end;

        bool found = false,
             first = true;
        int i;

        BitArray cip(8), tmp_cip(8), msg(8), key(10), tmp_key(10);
        cip.fromString(args.o_cip);
        msg.fromString(args.o_msg);
        key.fromString(args.o_key);

        gettimeofday(&t_start, NULL);
        for(i = 0; i<15; i++){
            std::string key_s = "0000000000";
            while(key_s != "1111111111"){
                tmp_key.fromString(key_s);
                tmp_cip = _crypt(M_ENCRYPT,msg,tmp_key);
                if(tmp_cip.toString() == cip.toString()){
                    found = true;
                    if(i == 0){
                        key_list.push_back(tmp_key.toString());
                    }
                }
                key_s = b_add(key_s);
            }
        }
        gettimeofday(&t_end, NULL);
        std::cout << "\nMessage:\t" << args.o_msg << ":" << string2int(args.o_msg) << std::endl;
        std::cout << "Cipher:\t\t" << cip.toString() << ":" << string2int(cip.toString()) << std::endl;


        if(found){
            std::cout << "key(s):\t\t";
            for(it = key_list.begin(); it != key_list.end(); it++){
                if(first){
                    std::cout << *it << ":" << string2int(*it) << std::endl;
                    first = false;
                }
                else
                    std::cout << "\t\t" << *it << ":" << string2int(*it) << std::endl;
            }
        }
        else
            std::cout << "key(s):\t\tnot found!" << std::endl;
        double tS = t_start.tv_sec*1000000 + (t_start.tv_usec);
        double tE = t_end.tv_sec*1000000  + (t_end.tv_usec);
        std::cout << "\nTime:\t\t" << ((tE - tS)/15)/1000000 << "(s)" << std::endl;
    }
    else{
        print_usage();
    }
    return 0;
}
